import autotda.functions.tree_functions as tree_functions
import autotda.functions.evaluation_functions as evaluation_functions


def show_features(autofeat, tree_id: int, show_discarded_features: bool = False):
    tree = tree_functions.get_tree_by_id(autofeat, tree_id)
    tree.show_table(show_discarded_features)


def adjust_relevance_value(autofeat, tree_id, feature_name, value):
    tree = tree_functions.get_tree_by_id(autofeat, tree_id)
    if tree is None:
        return
    for join in tree.joins:
        for index, item in enumerate(join.rel_red["rel"]):
            if item[0] == feature_name:
                new_item = (item[0], value)
                join.rel_red["rel"][index] = new_item
    evaluation_functions.rerun(autofeat)


def adjust_redundancy_value(autofeat, tree_id, feature_name, value):
    tree = tree_functions.get_tree_by_id(autofeat, tree_id=tree_id)
    if tree is None:
        return
    for join in tree.joins:
        for index, item in enumerate(join.rel_red["red"]):
            if item[0] == feature_name:
                new_item = (item[0], value)
                join.rel_red["red"][index] = new_item
    evaluation_functions.rerun(autofeat)


def adjust_non_null_ratio(autofeat, tree_id, table_name, value):
    tree = tree_functions.get_tree_by_id(autofeat, tree_id)
    if tree is None:
        return
    for join in tree.joins:
        if join.to_table == table_name:
            join.non_null_ratio = value
    evaluation_functions.rerun(autofeat)


def move_features_to_discarded(autofeat, tree_id, features):
    tree = tree_functions.get_tree_by_id(autofeat, tree)
    if tree is None:
        return
    for join in tree.joins:
        for index, item in enumerate(join.rel_red["rel"]):
            if item[0] in features:
                join.rel_red["rel"].pop(index)
                join.rel_red_discarded["rel"].append(item)
        for index, item in enumerate(join.rel_red["red"]):
            if item[0] in features:
                join.rel_red["red"].pop(index)
                join.rel_red_discarded["red"].append(item)
    new_features = set(tree.features).difference(set(features))
    tree.features = list(new_features)
    # evaluation_functions.rerun(autofeat)


def move_features_to_selected(autofeat, tree_id, features):
    tree = tree_functions.get_tree_by_id(autofeat, tree_id)
    if tree is None:
        return
    for join in tree.joins:
        for index, item in enumerate(join.rel_red_discarded["rel"]):
            if item[0] in features:
                join.rel_red_discarded["rel"].pop(index)
                join.rel_red["rel"].append(item)
        for index, item in enumerate(join.rel_red_discarded["red"]):
            if item[0] in features:
                join.rel_red_discarded["red"].pop(index)
                join.rel_red["red"].append(item)
    tree.features.extend(features)
    # evaluation_functions.rerun(autofeat)



import os

from autotda.functions.classes import Weight
import pandas as pd
from multiprocessing import Manager
from joblib import Parallel, delayed
import itertools
import tqdm
import seaborn
from valentine.algorithms import Coma, JaccardLevenMatcher
from valentine import valentine_match
import matplotlib.pyplot as plt
from tabulate import tabulate
import autotda.functions.tree_functions as tree_functions


def read_relationships(self, file_path):
    self.weights = []
    f = open(file_path, "r")
    stringlist = f.read().split(",")
    for i in stringlist:
        if i != "":
            table1, table2, col1, col2, weight = i.split("--")
            self.weights.append(Weight(table1, table2, col1, col2, float(weight)))
    f.close()
    tables = self.get_tables_repository()
    self.weight_string_mapping = {}
    for t in tables:
        if len(t) > 20:
            new_string = t.split("/")[0] + "/" + t.split("/")[1][:3] + "..." + t.split("/")[1][-7:]
            self.weight_string_mapping[t] = new_string
        else:
            self.weight_string_mapping[t] = t


def find_relationships(autofeat, relationship_threshold: float = 0.5, matcher: str = "coma", 
                       explain=False, verbose=True, use_cache=True):

    tables = autofeat.get_tables_repository()
    tables.extend(autofeat.extra_tables)
    tables = [i for i in tables if i not in autofeat.exclude_tables]
    if explain:
        print_string = f" 1. AutoFeat computes the relationships between {len(tables)} tables from the datasets: {autofeat.datasets} repository using {matcher} similarity score with a threshold of {relationship_threshold} (i.e., all the relationships with a similarity < {relationship_threshold} will be discarded) " 
        if len(autofeat.extra_tables) > 0:
            print_string += f"using extra tables: {autofeat.extra_tables} "
        if len(autofeat.exclude_tables) > 0:
            print_string += f"and excludes tables: {autofeat.exclude_tables}"
        print_string += f""
        print(print_string)
    if verbose:
        print("Calculating relationships...")

    filename = f"saved_weights/{autofeat.base_table}_{relationship_threshold}_{matcher}_weights.txt"
    if os.path.isfile(filename):
        read_relationships(autofeat, filename)
        return

    manager = Manager()
    temp = manager.list()
    autofeat.relationship_threshold = relationship_threshold
    autofeat.matcher = matcher
    
    # This function calculates the COMA weights between 2 tables in the datasets.
    def calculate_matches(table1: pd.DataFrame, table2: pd.DataFrame, matcher: str) -> dict:
        if matcher == "jaccard":
            matches = valentine_match(table1, table2, JaccardLevenMatcher())
        else:
            matches = valentine_match(table1, table2, Coma())
        return matches
    
    def profile(combination, matcher="coma"):
        (table1, table2) = combination
        df1 = pd.read_csv("data/benchmark/" + table1)
        df2 = pd.read_csv("data/benchmark/" + table2)
        matches = calculate_matches(df1, df2, matcher)
        for m in matches.items():
            ((_, col_from), (_, col_to)), similarity = m
            if similarity > relationship_threshold:
                temp.append(Weight(table1, table2, col_from, col_to, similarity))
                temp.append(Weight(table2, table1, col_to, col_from, similarity))
    autofeat.weight_string_mapping = {}
    for t in tables:
        if len(t) > 20:
            new_string = t.split("/")[0] + "/" + t.split("/")[1][:3] + "..." + t.split("/")[1][-7:]
            autofeat.weight_string_mapping[t] = new_string
        else:
            autofeat.weight_string_mapping[t] = t

    Parallel(n_jobs=-1)(delayed(profile)(combination, matcher)
                        for combination in tqdm.tqdm(itertools.combinations(tables, 2), 
                                                     total=len(tables) * (len(tables) - 1) / 2))
    autofeat.weights = temp

    if use_cache:
        os.mkdir("saved_weights/" + autofeat.base_table.split("/")[0])
        f = open(f"saved_weights/{autofeat.base_table}_{relationship_threshold}_{matcher}_weights.txt", "w")
        stringlist = []
        for i in autofeat.weights:
            stringlist.append(f"{i.from_table}--{i.to_table}--{i.from_col}--{i.to_col}--{i.weight},")
        f.writelines(stringlist)
        f.close()


def add_relationship(autofeat, table1: str, col1: str, table2: str, col2: str, weight: float, update: bool = True):
    autofeat.weights.append(Weight(table1, table2, col1, col2, weight))
    autofeat.weights.append(Weight(table2, table1, col2, col1, weight))
    if update:
        tree_functions.rerun(autofeat)


def remove_relationship(autofeat, table1: str, col1: str, table2: str, col2, update: bool = True):
    weights = [i for i in autofeat.weights if i.from_table == table1 
               and i.to_table == table2 and i.from_col == col1 
               and i.to_col == col2]
    weights = weights + [i for i in autofeat.weights if i.from_table == table2 and i.to_table == table1 
                         and i.from_col == col2 and i.to_col == col1]
    if len(weights) == 0:
        return
    for i in weights:
        if i in autofeat.weights:
            autofeat.weights.remove(i)
    if update:
        tree_functions.rerun(autofeat)


def update_relationship(autofeat, table1: str, col1: str, table2: str, col2: str, weight: float):
    remove_relationship(autofeat, table1, col1, table2, col2, update=False)
    add_relationship(autofeat, table1, col1, table2, col2, weight, update=False)
    tree_functions.rerun(autofeat)


def display_best_relationships(autofeat):
    tables = autofeat.get_tables_repository()
    highest_weights = []
    for table1 in tables:
        for table2 in tables:
            if table1 == table2:
                highest_weights.append([autofeat.weight_string_mapping[table1], 
                                        autofeat.weight_string_mapping[table2], 1])
            else:
                weight = get_best_weight(autofeat, table1, table2)
                if weight is not None:
                    highest_weights.append([autofeat.weight_string_mapping[weight.from_table], 
                                            autofeat.weight_string_mapping[weight.to_table], 
                                            weight.weight])
    if len(autofeat.datasets) == 1:
        highest_weights = [[i[0].split("/")[-1], i[1].split("/")[-1], i[2]] for i in highest_weights]
    df = pd.DataFrame(highest_weights, columns=["from_table", "to_table", "weight"])
    seaborn.heatmap(df.pivot(index="from_table", columns="to_table", values="weight"), square=True, cmap="PiYG",
                    vmin=0, vmax=1)
    plt.xlabel("")
    plt.ylabel("")
    plt.xticks(fontsize="small", rotation=60)
    plt.yticks(fontsize="small", rotation=0)
    plt.savefig("heatmap.pdf", dpi=300, bbox_inches='tight')


def display_table_relationship(autofeat, table1: str, table2: str):
    weights = [i for i in autofeat.weights if i.from_table == table1 and i.to_table == table2]
    if len(weights) == 0:
        print("No Weights found")
        return
    df = pd.DataFrame([[i.from_col, i.to_col, i.weight] for i in weights], 
                      columns=["from_column", "to_column", "weight"])
    plt.figure(figsize=(15, 3))
    seaborn.heatmap(df.pivot(index="from_column", columns="to_column", values="weight"), square=True, cmap="PiYG", 
                    vmin=0, vmax=1)
    plt.xlabel(table2)
    plt.ylabel(table1)
    tab1_name = table1.split('/')[-1].split('.')[0]
    tab2_name = table2.split('/')[-1].split('.')[0]
    plt.xticks(fontsize="small", rotation=60)
    plt.yticks(fontsize="small", rotation=0)
    plt.savefig(f"heatmap_{tab1_name}_{tab2_name}.pdf", dpi=300, bbox_inches='tight')


def explain_relationship(autofeat, table1: str, table2: str):
    weights = [i for i in autofeat.weights if (i.from_table == table1 or i.from_table == table2) 
               and (i.to_table == table1 or i.to_table == table2)]
    rows = []
    if len(weights) == 0:
        print(f"There are no relationships between {table1} and {table2}.")
        return
    for i in weights:
        rows.append([autofeat.weight_string_mapping[i.from_table], 
                     autofeat.weight_string_mapping[i.to_table], i.weight])
    print(f"Relationships between {table1} and {table2}:")
    table = tabulate(rows, headers=["from_table", "to_table", "weight"])
    print(table)
    

def get_best_weight(autofeat, table1: str, table2: str) -> Weight:
    weights = [i for i in autofeat.weights if i.from_table == table1 and i.to_table == table2]
    if len(weights) == 0:
        return None
    return max(weights, key=lambda x: x.weight)


def rerun(autofeat, threshold, matcher):
    if len(autofeat.weights) > 0:
        print("Relationships are recaclculated.")
        find_relationships(autofeat, threshold, matcher)
        tree_functions.rerun(autofeat)



from autotda.functions.classes import Tree, Join, Weight
import networkx
from matplotlib import pyplot as plt
from pathlib import Path as pt
import pandas as pd
from copy import deepcopy
from autotda.functions.helper_functions import get_df_with_prefix
import uuid
import datetime
from typing import Tuple, Optional
from autogluon.features.generators import AutoMLPipelineFeatureGenerator
import autotda.functions.evaluation_functions as evaluation_functions
import autotda.functions.feature_functions as feature_functions
import autotda.functions.evaluation_functions as evaluation_functions
from networkx.drawing.nx_pydot import graphviz_layout


def compute_join_trees(autofeat, top_k_features, non_null_ratio_threshold: float = 0.5, explain=False, verbose=True):
    autofeat.trees = []
    autofeat.discovered = set()
    autofeat.join_name_mapping = {}
    autofeat.set_base_table(autofeat.base_table, autofeat.targetColumn)
    autofeat.top_k_features = top_k_features
    emptyTree = Tree(begin=autofeat.base_table, joins=[], rank=0)
    emptyTree.features = autofeat.partial_join_selected_features[str(autofeat.base_table)]
    autofeat.tree_hash[str(autofeat.base_table)] = emptyTree
    emptyTree.id = 0
    autofeat.trees.append(emptyTree)
    if verbose:
        print("Calculating join trees...")
    stream_feature_selection(autofeat=autofeat, top_k_features=top_k_features, queue={str(autofeat.base_table)},
                             non_null_ratio_threshold=non_null_ratio_threshold)


def stream_feature_selection(autofeat, top_k_features, queue: set, non_null_ratio_threshold: float,
                             previous_queue: set = None):
    if len(queue) == 0:
        return
    if previous_queue is None:
        previous_queue = queue.copy()

    all_neighbours = set()
    while len(queue) > 0:
        base_node_id = queue.pop()
        autofeat.discovered.add(base_node_id)
        near_weights = get_adjacent_nodes(autofeat, {base_node_id}, 0.5)
        neighbours = [n.to_table for n in near_weights]
        neighbours = sorted((set(neighbours).difference(autofeat.discovered)))
        all_neighbours.update(neighbours)
        for n in neighbours:
            autofeat.discovered.add(n)
            join_keys = autofeat.get_weights_from_and_to_table(base_node_id, n)
            max_val = 0
            highest_join_keys = []
            for jk in join_keys:
                if jk.weight > max_val:
                    max_val = jk.weight
                    highest_join_keys = [jk]
                elif jk.weight == max_val:
                    highest_join_keys.append(jk)
            right_df = get_df_with_prefix(autofeat.repository_location, highest_join_keys[0].to_table)
            current_queue = set()
            while len(previous_queue) > 0:
                previous_table_join: [Join] = previous_queue.pop()
                previous_join = None
                if previous_table_join == str(autofeat.base_table):
                    previous_join = autofeat.partial_join.copy()
                else:
                    read_file = autofeat.join_name_mapping[str(previous_table_join)]
                    key_path = pt(autofeat.temp_dir.name) / read_file
                    previous_join = pd.read_parquet(key_path)
                prop: Weight
                for prop in highest_join_keys:
                    join_list: [str] = previous_table_join + "--" + prop.to_table
                    filename = f"{autofeat.base_table.replace('/', '-')}_{str(uuid.uuid4())}.parquet"
                    sampled_right_df = right_df.groupby(prop.get_to_prefix()).sample(n=1, random_state=42)
                    if previous_join[prop.get_from_prefix()].dtype != sampled_right_df[prop.get_to_prefix()].dtype:
                        current_queue.append(previous_table_join)
                        continue
                    joined_df = pd.merge(left=previous_join, right=sampled_right_df, left_on=(prop.get_from_prefix()),
                                         right_on=(prop.get_to_prefix()), how="left")
                    joined_df.to_parquet(pt(autofeat.temp_dir.name) / filename)
                    non_null_ratio = non_null_ratio_calculation(joined_df, prop)
                    if non_null_ratio < non_null_ratio_threshold:
                        current_queue.add(previous_table_join)
                        continue
                    result = streaming_relevance_redundancy(
                        autofeat,
                        top_k_features=top_k_features,
                        dataframe=joined_df.copy(),
                        new_features=list(right_df.columns),
                        selected_features=autofeat.partial_join_selected_features[previous_table_join],
                    )
                    if result is not None:
                        score, rel_score, red_score, final_features, rel_discarded, red_discarded = result
                        remaining_rel_score = [i for i in rel_score if i[0] in final_features]
                        remaining_red_score = [i for i in red_score if i[0] in final_features]
                        red_discarded = red_discarded + [i for i in red_score if i[0] not in final_features]
                        rel_discarded = rel_discarded + [i for i in rel_score if i[0] not in final_features]
                        join = Join(prop.from_table, prop.to_table, 
                                    prop.from_col, prop.to_col, non_null_ratio, {"rel": remaining_rel_score, 
                                                                             "red": remaining_red_score}, 
                                    {"rel": rel_discarded, "red": red_discarded})
                        all_features = autofeat.partial_join_selected_features[str(previous_table_join)].copy()

                        all_features.extend(final_features)
                        autofeat.partial_join_selected_features[str(join_list)] = all_features
                        tree = deepcopy(autofeat.tree_hash[str(previous_table_join)])
                        tree.features = all_features

                        join_keys = autofeat.join_keys[str(previous_table_join)]
                        join_keys.extend([prop.get_from_prefix(), prop.get_to_prefix()])
                        autofeat.join_keys[str(join_list)] = join_keys
                        tree.join_keys = join_keys
                        tree.rank = score
                        tree.add_join(join)
                        tree.id = len(autofeat.trees)
                        autofeat.tree_hash[str(join_list)] = tree
                        autofeat.trees.append(deepcopy(tree))
                    else:
                        autofeat.partial_join_selected_features[str(join_list)] = \
                            autofeat.partial_join_selected_features[str(previous_table_join)]
                        autofeat.tree_hash[str(join_list)] = autofeat.tree_hash[str(previous_table_join)]
                        autofeat.join_keys[str(join_list)] = autofeat.join_keys[str(previous_table_join)]
                    autofeat.join_name_mapping[str(join_list)] = filename
                    current_queue.add(join_list)
            previous_queue.update(current_queue)
            
    stream_feature_selection(autofeat, top_k_features, all_neighbours, non_null_ratio_threshold, previous_queue)


def display_join_trees(self, top_k: None):
    if top_k is None:
        top_k = len(self.trees)
    sorted_trees = sorted(self.trees, key=lambda x: x.rank, reverse=True)[:top_k]
    for index, tree in enumerate(sorted_trees):
        if len(tree.joins) > 0:
            plt.figure(figsize=(5, 2))
            graph = networkx.DiGraph()
            plt.gca()
            labels = {}
            mapping = {tree.joins[0].from_table: 0}
            count = 0
            for i in tree.joins:
                count += 1
                mapping[i.to_table] = count
                graph.add_edge(mapping[i.from_table], mapping[i.to_table])
                from_col = i.from_col if len(i.from_col) < 20 else i.from_col[:10] + "..." + i.from_col[-10:]
                to_col = i.to_col if len(i.to_col) < 20 else i.to_col[:10] + "..." + i.to_col[-10:]
                labels[(mapping[i.from_table], mapping[i.to_table])] = (from_col + " -> " + to_col)
            ids = list(mapping.values())
            names = list(mapping.keys())
            df = pd.DataFrame({"Node ID": ids, "Table Name": names})
            pos = graphviz_layout(graph, prog="dot")
            networkx.draw(graph, pos=pos, with_labels=True)
            # networkx.draw_networkx_edge_labels(graph, pos=pos, edge_labels=labels, font_size=10)
            plt.title(f"Join Tree ID: {tree.id}. Rank: {('%.2f' % tree.rank)}")
            plt.table(cellText=df.values, cellLoc="center", colLabels=df.columns, loc="right").auto_set_column_width([0,1])
            plt.show()


def display_join_tree(self, tree_id):
    tree = get_tree_by_id(self, tree_id)
    if tree is None or len(tree.joins) == 0:
        return
    graph = networkx.DiGraph()
    plt.gca()
    labels = {}
    mapping = {tree.joins[0].from_table: 0}
    count = 0
    for i in tree.joins:
        count += 1
        mapping[i.to_table] = count
        graph.add_edge(mapping[i.from_table], mapping[i.to_table])
        from_col = i.from_col if len(i.from_col) < 20 else i.from_col[:10] + "..." + i.from_col[-10:]
        to_col = i.to_col if len(i.to_col) < 20 else i.to_col[:10] + "..." + i.to_col[-10:]
        labels[(mapping[i.from_table], mapping[i.to_table])] = (from_col + " -> " + to_col)
    ids = list(mapping.values())
    names = list(mapping.keys())
    df = pd.DataFrame({"Node ID": ids, "Feature Name": names})
    pos = graphviz_layout(graph, prog="dot")
    networkx.draw(graph, pos=pos, with_labels=True)
    # networkx.draw_networkx_edge_labels(graph, pos=pos, edge_labels=labels, font_size=10)
    plt.title(f"Join Tree ID: {tree.id}. Rank: {('%.2f' % tree.rank)}")
    plt.table(cellText=df.values, cellLoc="center", colLabels=df.columns, loc="right").auto_set_column_width([0, 1])
    plt.show()


def streaming_relevance_redundancy(
    self, top_k_features, dataframe: pd.DataFrame, new_features: list[str], selected_features: list[str]
) -> Optional[Tuple[float, list[dict]]]:
    t = datetime.datetime.now()
    df = AutoMLPipelineFeatureGenerator(
        enable_text_special_features=False, enable_text_ngram_features=False
    ).fit_transform(X=dataframe, random_state=42, random_seed=42)
    X = df.drop(columns=[self.targetColumn])
    y = df[self.targetColumn]

    features = list(set(X.columns).intersection(set(new_features)))
    top_feat = len(features) if len(features) < top_k_features else top_k_features

    m = 1
    feature_score_relevance, rel_discarded_features = self.rel_red.measure_relevance(
        dataframe=X, new_features=features, target_column=y
    )
    feature_score_relevance = feature_score_relevance[:top_feat]
    rel_discarded_features += feature_score_relevance[top_feat:]

    if len(feature_score_relevance) == 0:
        return None
    relevant_features = list(dict(feature_score_relevance).keys())
    m = len(feature_score_relevance) if len(feature_score_relevance) > 0 else m
    sum_m = sum(list(map(lambda x: x[1], feature_score_relevance)))
    final_features = relevant_features
    sum_o = 0
    o = 1
    feature_score_redundancy, red_discarded_features = self.rel_red.measure_redundancy(
        dataframe=X, selected_features=selected_features, relevant_features=relevant_features, target_column=y
    )

    if len(feature_score_redundancy) == 0:
        return None

    o = len(feature_score_redundancy) if feature_score_redundancy else o
    sum_o = sum(list(map(lambda x: x[1], feature_score_redundancy)))
    final_features = list(dict(feature_score_redundancy).keys())

    score = (o * sum_m + m * sum_o) / (m * o)

    return (score, feature_score_relevance, feature_score_redundancy, final_features, 
            rel_discarded_features, red_discarded_features)


def inspect_join_tree(self, tree_id):
    tree = get_tree_by_id(self, tree_id)
    if tree is None:
        return
    print(tree)
    feature_functions.show_features(self, tree_id)


def materialise_join_tree(self, tree_id):
    tree = get_tree_by_id(self, tree_id)
    base_df = get_df_with_prefix(self.repository_location, self.base_table, self.targetColumn)
    i: Join
    for i in tree.joins:
        df = get_df_with_prefix(self.repository_locatoin, i.to_table)
        base_df = pd.merge(base_df, df, left_on=i.get_from_prefix(), right_on=i.get_to_prefix(), how="left")
    base_df = base_df[tree.features]
    # Filter on selected features in rel_red
    return base_df


def get_tree_by_id(self, tree_id) -> Tree:
    list = [i for i in self.trees if i.id == tree_id]
    if len(list) == 0:
        return None
    return list[0]


def get_adjacent_nodes(self, nodes: list, threshold: float) -> [Weight]:
    return_list = list()
    for n in nodes:
        for x in self.get_weights_from_table(n):
            return_list.append(x)
    return return_list


def non_null_ratio_calculation(joined_df: pd.DataFrame, prop: Weight) -> float:
    total_length = joined_df.shape[0]
    non_nulls = joined_df[prop.get_to_prefix()].count()
    return non_nulls / total_length


def remove_join_from_tree(self, tree_id: int, table: str):
    tree = get_tree_by_id(self, tree_id)
    if tree is None:
        print("Can not find path")
        return
    for index, join in enumerate(tree.joins):
        if join.to_table == table:
            features = [i[0] for i in join.rel_red["rel"]] + [i[0] for i in join.rel_red["red"]]
            tree.features = list(set(tree.features).difference(set(features)))
            tree.joins.pop(index)
            print("Path removed")
    # evaluation_functions.rerun(self)


def rerun(autofeat):
    if len(autofeat.trees) > 0:
        print("Recalculating trees")
        compute_join_trees(autofeat, top_k_features=autofeat.top_k_features)
        evaluation_functions.rerun(autofeat)


def explain_tree(self, tree_id: int):
    tree = get_tree_by_id(self, tree_id)
    if tree is None:
        return
    print(tree.explain())


"""
This file contains the classes for the AutoFeatInsights package.
"""
import pandas as pd
import networkx as nx
from tabulate import tabulate
import matplotlib.pyplot as plt


class Join():

    def __init__(self, from_table: str,
                 to_table: str, from_col: str, to_col: str,
                 non_null_ratio: float, rel_red: dict, rel_red_discarded: dict):

        self.from_table = from_table
        self.to_table = to_table
        self.from_col = from_col
        self.to_col = to_col
        self.non_null_ratio = non_null_ratio
        self.rel_red = rel_red
        self.rel_red_discarded = rel_red_discarded
   
    def get_from_prefix(self):
        return self.from_table + "." + self.from_col
  
    def get_to_prefix(self):
        return self.to_table + "." + self.to_col
   
    def __str__(self) -> str:
        return "Join from " + self.from_table + "." + self.from_col + " to " \
            + self.to_table + "." + self.to_col + " with Non-Null ratio " \
            + str(self.non_null_ratio) + "."

    def __repr__(self) -> str:
        return f"{self.from_table}.{self.from_col} -> {self.to_table}.{self.to_col}"
 
    def explain(self) -> str:
        return "This is a join from: " + self.get_from_prefix() \
            + " to " + self.get_to_prefix() \
            + " with Non-Null Ratio: " + str(self.nnon_ull_ratio) + "." 


class Tree:

    features: list = []
    join_keys: list = []

    def __init__(self, begin: str, joins: [Join] = None, rank: float = None):
        self.begin = begin
        if joins is None:
            self.joins = []
        else:
            self.joins = joins
        self.rank = rank
    
    def add_join(self, join):
        self.joins.append(join)

    def get_rel_red(self):
        rel_rel_dict = {}
        i: Join
        for i in self.joins:
            # dict[i.to_table + "." + i.to_col] = {}
            for rel_red in i.rel_red:
                for j in i.rel_red[rel_red]:
                    name, val = j
                    if name not in rel_rel_dict:
                        rel_rel_dict[name] = {"non_null_ratio": 0, "rel": None, "red": None}
                    rel_rel_dict[name][rel_red] = val
                    if rel_rel_dict[name]["non_null_ratio"] == 0:
                        rel_rel_dict[name]["non_null_ratio"] = i.non_null_ratio
        return rel_rel_dict
    
    def get_discarded_rel_red(self):
        rel_rel_dict = {}
        i: Join
        for i in self.joins:
            # dict[i.to_table + "." + i.to_col] = {}
            for rel_red in i.rel_red_discarded:
                for j in i.rel_red_discarded[rel_red]:
                    name, val = j
                    if name not in rel_rel_dict:
                        rel_rel_dict[name] = {"non_null_ratio": 0, "rel": None, "red": None}
                    rel_rel_dict[name][rel_red] = val
                    if rel_rel_dict[name]["non_null_ratio"] == 0:
                        rel_rel_dict[name]["non_null_ratio"] = i.non_null_ratio
        return rel_rel_dict
    
    def show_table(self, discarded_features: bool = False):
        scores = self.get_rel_red()
        table_data = []
        for key, values in scores.items():
            # if len(key) > 20:
            #     key = key[:10] + "..." + key[-10:]
            row = [key, values["non_null_ratio"], values['rel'], values['red']]
            table_data.append(row)
        print('\n\tSelected features:\n')
        # Displaying the table 
        table = tabulate(table_data, headers=["Key", "Non-Null Ratio", "Relevance", "Redundancy"], tablefmt="grid")
        print(table)
        if discarded_features:
            print("Discarded Features")
            discarded_scores = self.get_discarded_rel_red()
            table_data = []
            for key, values in discarded_scores.items():
                if len(key) > 20:
                    key = key[:10] + "..." + key[-10:]
                row = [key, values["non_null_ratio"], values['rel'], values['red']]
                table_data.append(row)
            # Displaying the table
            table = tabulate(table_data, headers=["Key", "Non-Null Ratio", "Relevance", "Redundancy"], tablefmt="grid")
            print(table)

    def __str__(self) -> str:
        ret_string = "Rank: " + ("%.2f" % self.rank)
        ret_string += "\nFrom base table: " + self.begin
        table = tabulate([[i.from_table + "." + i.from_col, i.to_table + "." + i.to_col, i.non_null_ratio] 
                          for i in self.joins], headers=["From Table.Column", "To Table.Column", "Non-Null Ratio"], tablefmt="grid")
        ret_string += "\nJoin Trees: \n" + table
        return ret_string
    
    def __repr__(self) -> str:
        return self.begin + " -> " + str(self.joins)
   
    def explain(self) -> str:
        ret_str = "The join tree starts at the table " + self.begin \
            + ". \nThe join tree has the following joins: "
        for i in self.joins:
            ret_str += "\n \t from (table.column) " + i.get_from_prefix() \
                + " to (table.column) " + i.get_to_prefix() + " with Non-Null ratio " \
                + ("%.2f" % i.non_null_ratio) + "."
        ret_str += "\nThe rank of the tree is " + ("%.2f" % self.rank) + "."
        return ret_str
    

class Result():

    # rank: float
    tree: Tree
    accuracy: float
    feature_importance: dict
    model: str
    model_full_name: str
    data: pd.DataFrame

    def __init__(self):
        self.feature_importance = {}

    def getFeatureImportance(self, tableName=None, featureName=None):
        if tableName is None and featureName is None:
            return self.result.feature_importance
        elif tableName is not None and featureName is None:
            return self.getFeatureImportanceByTable(self, tableName)
        elif tableName is not None and featureName is not None:
            return self.getFeatureImportanceByTableAndFeature(self, tableName,
                                                              featureName)
    
    def getFeatureImportanceByTable(self, tableName: str):
        list = []
        for i in self.feature_importance:
            if tableName in i:
                list.append(i)
        return list
    
    def getFeatureImportanceByTableAndFeature(self, tableName: str,
                                              featureName: str):
        return self.feature_importance[tableName + "." + featureName]
    
    def show_graph(self, ax=None, plot=True):
        G = nx.Graph()
        if len(self.tree.joins) == 0:
            G.add_node(self.tree.begin)
        else:
            for i in self.tree.joins:
                G.add_edge(i.from_table, i.to_table)
        if plot:
            plt.figure()
            plt.title(self.model + " : " + str(self.accuracy) + " accuracy")
        if ax is None:
            nx.draw(G, with_labels=True, font_weight='bold')
        else:
            nx.draw(G, ax=ax, with_labels=True, font_weight='bold')
            ax.set_title(self.model + ": " + str(self.accuracy) + " accuracy")
        if plot:
            plt.show()

    def __str__(self) -> str:
        ret_string = "Result with model " + self.model_full_name \
            + "\n \t rank:" + ("%.2f" % self.rank) \
            + " \n \t with tree: \n" + str(self.tree) \
            + " \n \t Accuracy:" + ("%.2f" % self.accuracy) \
            + " \n \t Feature importance:"
        for i, j in self.feature_importance.items():
            ret_string += f"\n\t\t{str(i)} : {str(j)}"
        return ret_string
    
    def explain(self) -> str:
        feature_list = list(map(lambda x: [x[0], x[1]], self.feature_importance.items()))
        # for i in self.feature_importance[0]:
        #     if len(i) > 20:
        #         i_string = i[:10] + "..." + i[-10:]
        #     else:
        #         i_string = i
        #     feature_list.append([i_string, self.feature_importance[0][i]])
        return f"The result is calculated by evaluating the dataset corresponding to the join tree\n" \
               f" using the AutoML framework AutoGluon." \
            + f"\n The accuracy of the model {self.model_full_name} is {self.accuracy}." \
            + "\n The model used the following features: \n" \
            + tabulate(feature_list, headers=["Feature", "Importance"], tablefmt="grid")


class Weight():
    from_table: str
    to_table: str
    from_col: str
    to_col: str
    weight: float

    def __init__(self, from_table, to_table, from_col, to_col, weight):
        self.from_table = from_table
        self.to_table = to_table
        self.from_col = from_col
        self.to_col = to_col
        self.weight = weight

    def get_from_prefix(self):
        return self.from_table + "." + self.from_col
    
    def get_to_prefix(self):
        return self.to_table + "." + self.to_col
    
    def __str__(self) -> str:
        return "Weight from " + self.get_from_prefix() \
            + " to " + self.get_to_prefix() \
            + " with weight " + str(self.weight)
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def explain(self) -> str:
        return "This weight is calculated by the COMA algorithm. \
            This calculates the similarity between the columns of the tables. \
            The higher the similarity, the higher the weight. \n \
            The weight from " + self.getFromPrefix() + " to " + self.getToPrefix() + " is " + str(self.weight) + "."



from autotda.functions.helper_functions import get_df_with_prefix
import tqdm
from autotda.functions.classes import Result, Join
import autotda.functions.tree_functions as tree_functions
from autogluon.features.generators import AutoMLPipelineFeatureGenerator
from autogluon.tabular import TabularPredictor
from sklearn.model_selection import train_test_split
import pandas as pd

hyper_parameters = [
    {"RF": {}},
    {"GBM": {}},
    {"XT": {}},
    {"XGB": {}},
    {'KNN': {}},
    {'LR': {'penalty': 'L1'}},
]


def get_hyperparameters(algorithm: str = None) -> list[dict]:
    if algorithm is None:
        return hyper_parameters

    if algorithm == 'LR':
        return [{'LR': {'penalty': 'L1'}}]
    model = {algorithm: {}}
    if model in hyper_parameters:
        return [model]
    else:
        raise Exception(
            "Unsupported algorithm. Choose one from the list: [RF, GBM, XT, XGB, KNN, LR]."
        )
    

def evalute_trees(autofeat, algorithm, top_k_results: int = 5,
                  explain=False, verbose=True):
    autofeat.results = []
    autofeat.algorithm = algorithm
    autofeat.top_k_results = top_k_results
    sorted_trees = sorted(autofeat.trees, key=lambda x: x.rank, reverse=True)[:top_k_results]
    if verbose:
        print("Evaluating join trees...")
    for tree in tqdm.tqdm(sorted_trees, total=len(sorted_trees)):
        evaluate_table(autofeat, algorithm, tree.id, verbose=verbose, multiple=True)
    if explain:
        best_result = sorted(autofeat.results, key=lambda x: x.accuracy, reverse=True)[0]
        print(f"AutoFeat creates {len(autofeat.trees)} join trees: the best performing join tree is tree: {best_result.tree.id}")
        tree = tree_functions.get_tree_by_id(autofeat, best_result.tree.id)
        print(tree.explain())
        autofeat.show_features(best_result.tree.id)
        print(best_result.explain())


def evaluate_table(autofeat, algorithm, tree_id: int, verbose=False, multiple=False):
    if not multiple:
        autofeat.results = []
    tree = tree_functions.get_tree_by_id(autofeat, tree_id)
    base_df = get_df_with_prefix(autofeat.repository_location, autofeat.base_table, autofeat.targetColumn)
    i: Join
    for i in tree.joins:
        df = get_df_with_prefix(autofeat.repository_location, i.to_table).groupby(i.get_to_prefix()).sample(n=1, random_state=42)
        base_df = pd.merge(base_df, df, left_on=i.get_from_prefix(), right_on=i.get_to_prefix(), how="left")
    base_df = base_df[tree.features + [autofeat.targetColumn]]
    columns_to_drop = set(base_df.columns).intersection(set(tree.join_keys))
    base_df.drop(columns=list(columns_to_drop), inplace=True)
    df = AutoMLPipelineFeatureGenerator(
        enable_text_special_features=False, enable_text_ngram_features=False, 
        verbosity=0).fit_transform(X=base_df)
    hyper_parameters = get_hyperparameters(algorithm)
    for hyperparam in hyper_parameters:
        X_train, X_test, y_train, y_test = train_test_split(df.drop(columns=[autofeat.targetColumn]),
                                                            df[autofeat.targetColumn], test_size=0.2, random_state=10)
        
        X_train[autofeat.targetColumn] = y_train
        X_test[autofeat.targetColumn] = y_test
        predictor = TabularPredictor(label=autofeat.targetColumn,
                                     problem_type="binary",
                                     verbosity=0,
                                     path="AutogluonModels/" + "models").fit(
                                         train_data=X_train, hyperparameters=hyperparam)
        model_names = predictor.get_model_names()
        for model in model_names[:-1]:
            result = Result()
            res = predictor.evaluate(X_test, model=model)
            result.accuracy = res['accuracy']
            ft_imp = predictor.feature_importance(data=X_test, model=model, feature_stage="original")
            result.feature_importance = dict(zip(list(ft_imp.index), ft_imp["importance"]))
            result.model = list(hyperparam.keys())[0]
            result.model_full_name = model
            result.rank = tree.rank
            result.tree = tree
            autofeat.results.append(result)


def add_result(self, result):
    self.results.append(result)


def show_result(self, id: str):
    return self.results[id].show_graph


def get_best_result(autofeat):
    return max(autofeat.results, key=lambda x: x.accuracy)


def rerun(autofeat):
    if len(autofeat.results) > 0:
        print("Recalculating results...")
        evalute_trees(autofeat, autofeat.algorithm, top_k_results=autofeat.top_k_results)


def explain_result(self, tree_id: int, model: str):
    for i in self.results:
        if i.tree.id == tree_id and i.model == model:
            print(i.explain())
            return
    print("no result found")



import xxhash
from typing import List, Tuple, Optional
import pandas as pd
import numpy as np
from ITMO_FS.filters.multivariate import CMIM, MRMR
from ITMO_FS.utils.information_theory import entropy, conditional_mutual_information
from sklearn.preprocessing import KBinsDiscretizer
from scipy.stats import rankdata


xxh = xxhash.xxh64(seed=42)


class RelevanceRedundancy:

    def __init__(self, target_column: str, jmi: bool = False, pearson: bool = False):
        self.target_column = target_column
        self.target_entropy = None
        self.dataframe_entropy = {}
        self.dataframe_conditional_entropy = {}
        self.jmi = jmi
        self.pearson = pearson

    def measure_relevance(
        self, dataframe: pd.DataFrame, new_features: List[str], target_column: pd.Series
    ) -> List[tuple]:

        if self.target_column in new_features:
            new_features.remove(self.target_column)

        new_common_features = list(set(dataframe.columns).intersection(set(new_features)))
        if len(new_common_features) == 0:
            return []

        if self.pearson:
            correlation_score = abs(
                pearson_correlation(np.array(dataframe[new_common_features]), np.array(target_column))
            )
        else:
            correlation_score = abs(
                spearman_correlation(np.array(dataframe[new_common_features]), np.array(target_column))
            )

        final_feature_scores_rel = []
        discarded_features = []
        for value, name in list(zip(correlation_score, new_common_features)):
            # value 0 means that the features are completely redundant
            # value 1 means that the features are perfectly correlated, which is still undesirable
            if 0 < value < 1:
                final_feature_scores_rel.append((name, value))
            else:
                discarded_features.append((name, value))

        if len(final_feature_scores_rel) == 0:
            return []

        return sorted(final_feature_scores_rel, key=lambda s: s[1], reverse=True), discarded_features

    def measure_redundancy(
        self,
        dataframe: pd.DataFrame,
        selected_features: List[str],
        relevant_features: List[str],
        target_column: pd.Series,
    ) -> List[tuple]:
        """
        Measure the redundancy of selected features with respect to the relevant features.

        Args:
            dataframe (pd.DataFrame): The input dataframe.
            selected_features (List[str]): The list of selected features.
            relevant_features (List[str]): The list of relevant features.
            target_column (pd.Series): The target column.

        Returns:
            List[tuple]: A list of tuples containing the name of the feature and its redundancy score.
        """
        selected_features_int = [i for i, value in enumerate(dataframe.columns) if value in selected_features]
        new_features_int = [i for i, value in enumerate(dataframe.columns) if value in relevant_features]

        est = KBinsDiscretizer(strategy='uniform', encode='ordinal', subsample=1000)
        try:
            discr_dataframe = est.fit_transform(dataframe)
        except ValueError:
            discr_dataframe = dataframe

        relevance = np.apply_along_axis(
            self.cached_mutual_information, 0, np.array(discr_dataframe)[:, np.array(new_features_int)], target_column
        )
        redundancy = np.vectorize(
            lambda free_feature: np.sum(
                np.apply_along_axis(
                    self.cached_mutual_information,
                    0,
                    np.array(discr_dataframe)[:, np.array(selected_features_int)],
                    np.array(discr_dataframe)[:, free_feature],
                )
            )
        )(new_features_int)

        if self.jmi:
            cond_dependency = np.vectorize(
                lambda free_feature: np.sum(
                    np.apply_along_axis(
                        self.cached_conditional_mutual_information,
                        0,
                        np.array(dataframe)[:, np.array(selected_features_int)],
                        np.array(dataframe)[:, free_feature],
                        target_column,
                    )
                )
            )(np.array(new_features_int))
            mrmr_scores = (
                relevance
                - (1 / np.array(selected_features_int).size) * redundancy
                + (1 / np.array(selected_features_int).size) * cond_dependency
            )
        else:
            mrmr_scores = relevance - (1 / np.array(selected_features_int).size) * redundancy

        if np.all(np.array(mrmr_scores) == 0):
            return []

        max_mrmr = np.max(mrmr_scores)
        min_mrmr = np.min(mrmr_scores)
        if max_mrmr == min_mrmr:
            normalised_scores = (mrmr_scores - min_mrmr) / min_mrmr
        else:
            normalised_scores = (mrmr_scores - min_mrmr) / (max_mrmr - min_mrmr)

        feature_scores = list(zip(np.array(dataframe.columns)[np.array(new_features_int)], normalised_scores))
        final_feature_scores_mrmr = [(name, value) for name, value in feature_scores if value > 0]
        discarded_red_scores = list(set(feature_scores) - set(final_feature_scores_mrmr))
        
        return sorted(final_feature_scores_mrmr, key=lambda s: s[1], reverse=True), discarded_red_scores

    def measure_relevance_and_redundancy(
        self, dataframe: pd.DataFrame, selected_features: List[str], new_features: List[str], target_column: pd.Series
    ) -> Tuple[list, list]:
        final_feature_scores_rel = self.measure_relevance(dataframe, new_features, target_column)
        final_feature_scores_mrmr = self.measure_redundancy(
            dataframe, selected_features, list(dict(final_feature_scores_rel).keys()), target_column
        )
        return final_feature_scores_rel, final_feature_scores_mrmr

    def cached_mutual_information(self, x, y):
        xxh.update(x)
        xxh.update(np.array(y))
        h = xxh.intdigest()
        xxh.reset()
        if h in self.dataframe_conditional_entropy:
            cond_entropy = self.dataframe_conditional_entropy[h]
        else:
            cond_entropy = conditional_entropy(x, np.array(y))
            self.dataframe_conditional_entropy[h] = cond_entropy

        xxh.update(np.array(y))
        # h_entropy = hash(str(y))
        h_entropy = xxh.intdigest()
        xxh.reset()

        if h_entropy in self.dataframe_entropy:
            entr = self.dataframe_entropy[h_entropy]
        else:
            entr = entropy(y)
            self.dataframe_entropy[h_entropy] = entr
        return entr - cond_entropy

    def cached_conditional_mutual_information(self, x, y, z):
        h1 = hash(str(list(zip(x, z))))
        if h1 in self.dataframe_entropy:
            entropy_xz = self.dataframe_entropy[h1]
        else:
            entropy_xz = entropy(list(zip(x, z)))
            self.dataframe_entropy[h1] = entropy_xz

        h2 = hash(str(list(zip(y, z))))
        if h2 in self.dataframe_entropy:
            entropy_yz = self.dataframe_entropy[h2]
        else:
            entropy_yz = entropy(list(zip(y, z)))
            self.dataframe_entropy[h2] = entropy_yz

        h3 = hash(str(list(zip(x, y, z))))
        if h3 in self.dataframe_entropy:
            entropy_xyz = self.dataframe_entropy[h3]
        else:
            entropy_xyz = entropy(list(zip(x, y, z)))
            self.dataframe_entropy[h3] = entropy_xyz

        h4 = hash(str(z))
        if h4 in self.dataframe_entropy:
            entropy_z = self.dataframe_entropy[h4]
        else:
            entropy_z = entropy(z)
            self.dataframe_entropy[h4] = entropy_z

        return entropy_xz + entropy_yz - entropy_xyz - entropy_z


def measure_relevance(
    dataframe: pd.DataFrame, feature_names: List[str], target_column: pd.Series
) -> Tuple[Optional[list], list]:
    common_features = list(set(dataframe.columns).intersection(set(feature_names)))

    if len(common_features) == 0:
        return None, []

    features = dataframe[common_features]
    # scores = information_gain(np.array(features), np.array(target_column)) / max(entropy(features),
    #                                                                              entropy(target_column))
    scores = abs(spearman_correlation(np.array(features), np.array(target_column)))

    final_feature_scores = []
    final_features = []
    for value, name in list(zip(scores, common_features)):
        # value 0 means that the features are completely redundant
        # value 1 means that the features are perfectly correlated, which is still undesirable
        if 0 < value < 1:
            final_feature_scores.append((name, value))
            final_features.append(name)

    return final_feature_scores, final_features


def measure_conditional_redundancy(
    dataframe: pd.DataFrame,
    selected_features: List[str],
    new_features: List[str],
    target_column: pd.Series,
    conditional_redundancy_threshold: float = 0.5,
) -> Tuple[Optional[list], list]:
    selected_features_int = [i for i, value in enumerate(dataframe.columns) if value in selected_features]
    new_features_int = [i for i, value in enumerate(dataframe.columns) if value in new_features]

    scores = CMIM(
        np.array(selected_features_int), np.array(new_features_int), np.array(dataframe), np.array(target_column)
    )

    if np.all(np.array(scores) == 0):
        return None, []
    # normalise
    normalised_scores = (scores - np.min(scores)) / (np.max(scores) - np.min(scores))
    feature_scores = list(zip(np.array(dataframe.columns)[np.array(new_features_int)], normalised_scores))
    final_feature_scores = [(name, value) for name, value in feature_scores if value > 0]
    final_feature_names = [feat for feat, _ in final_feature_scores]

    return final_feature_scores, final_feature_names


def measure_redundancy(dataframe, feature_group: List[str], target_column) -> Tuple[Optional[list], List[str]]:
    if len(feature_group) == 1 or len(feature_group) == 0:
        return None, feature_group

    scores = np.vectorize(
        lambda feature: np.mean(
            np.apply_along_axis(
                lambda x, y, z: conditional_mutual_information(y, z, x),
                0,
                np.array(dataframe[np.setdiff1d(feature_group, feature)]),
                np.array(dataframe[feature]),
                np.array(target_column),
            )
        )
    )(feature_group)
    feature_scores = list(zip(np.array(feature_group), scores))
    final_feature_scores = [(name, value) for name, value in feature_scores if value > 0]
    final_feature_names = [feat for feat, _ in final_feature_scores]

    return final_feature_scores, final_feature_names


def measure_joint_mutual_information(
    dataframe: pd.DataFrame, selected_features: List[str], new_features: List[str], target_column: pd.Series
) -> Tuple[Optional[list], list]:
    selected_features_int = [i for i, value in enumerate(dataframe.columns) if value in selected_features]
    new_features_int = [i for i, value in enumerate(dataframe.columns) if value in new_features]

    scores = MRMR(
        np.array(selected_features_int), np.array(new_features_int), np.array(dataframe), np.array(target_column)
    )

    if np.all(np.array(scores) == 0):
        return None, []

    if np.max(scores) == np.min(scores):
        normalised_scores = (scores - np.min(scores)) / np.max(scores)
    else:
        normalised_scores = (scores - np.min(scores)) / (np.max(scores) - np.min(scores))

    feature_scores = list(zip(np.array(dataframe.columns)[np.array(new_features_int)], normalised_scores))
    final_feature_scores = [(name, value) for name, value in feature_scores if value > 0]
    final_feature_names = [feat for feat, _ in final_feature_scores]

    return final_feature_scores, final_feature_names


def get_df_with_prefix(repository_location, node_id: str, targetColumn=None):
    prefix = node_id + "."
    if targetColumn:
        dataframe = pd.read_csv(repository_location + '/' + node_id, encoding="utf8", engine="python", quotechar='"', 
                                escapechar='\\').set_index(targetColumn) \
            .add_prefix(prefix).reset_index()
    else:
        dataframe = pd.read_csv(repository_location + '/' + node_id, encoding="utf8", engine="python", quotechar='"', 
                                escapechar='\\').add_prefix(prefix)
    return dataframe


def conditional_entropy(x_j: np.ndarray, y: np.ndarray) -> np.float64:
    # Find unique values and their corresponding indices
    unique_x, inverse_x = np.unique(x_j, return_inverse=True)
    unique_y, inverse_y = np.unique(y, return_inverse=True)

    # Create a contingency table
    contingency_table = np.zeros((len(unique_x), len(unique_y)), dtype=int)
    np.add.at(contingency_table, (inverse_x, inverse_y), 1)

    # Compute counts for each unique value in x_j
    counts_x = np.sum(contingency_table, axis=1)

    # Calculate the probabilities P(Y|X)
    prob_y_given_x = contingency_table / counts_x[:, None]

    # Compute entropy for each unique value in x_j
    part_entropies = -np.sum(elog(prob_y_given_x), axis=1)

    # Compute the weighted average of the part_entropies
    entropy = np.sum(counts_x / len(x_j) * part_entropies)

    return entropy


def elog(value: np.ndarray) -> np.ndarray:
    log_values = np.zeros_like(value)
    mask = value > 0
    log_values[mask] = np.log(value[mask])
    return value * log_values


def pearson_correlation(x, y):
    x_dev = x - np.mean(x, axis=0)
    y_dev = y - np.mean(y)
    sq_dev_x = x_dev * x_dev
    sq_dev_y = y_dev * y_dev
    sum_dev = y_dev.T.dot(x_dev).reshape((x.shape[1],))
    denominators = np.sqrt(np.sum(sq_dev_y) * np.sum(sq_dev_x, axis=0))

    results = np.array(
        [(sum_dev[i] / denominators[i]) if denominators[i] > 0.0 else 0 for i
         in range(len(denominators))])
    return results


def spearman_correlation(x, y):
    n = x.shape[0]
    if n < 2:
        raise ValueError("The input should contain more than 1 sample")

    x_ranks = np.apply_along_axis(rankdata, 0, x)
    y_ranks = rankdata(y)

    return pearson_correlation(x_ranks, y_ranks)



from distutils.core import setup

setup(
  name = 'autotda',         # How you named your package folder (MyLib)
  packages = ['autotda', 'autotda.functions'],   # Chose the same as "name"
  version = '0.0.15',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Human in the loop Data Augmentation',   # Give a short description about your library
  author = 'Zeger Mouw',                   # Type in your name
  author_email = 'z.f.mouw@student.tudelft.nl',      # Type in your E-Mail
  url = 'https://github.com/delftdata/hci-auto-feat',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/delftdata/hci-auto-feat/archive/refs/tags/0.0.1.tar.gz',    # I explain this later on
  keywords = ['SOME', 'MEANINGFULL', 'KEYWORDS'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'autogluon',
        "deprecation",
        "dsplot",
        "graphviz",
        "ITMO-FS",
        "joblib",
        "matplotlib", 
        "neo4j",
        "networkx",
        "notebook",
        "numpy",
        "pandas",
        "pydot",
        "pyvis",
        "qpsolvers[daqp]",
        "scikit-learn",
        "seaborn",
        "tqdm",
        "typer",
        "valentine == 0.1.6"
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3.10',      #Specify which pyhton versions that you want to support
  ],
)